{*******************************************************}
{File:      NCOciFilter.PAS                             }
{Revision:  0.03.00 / 06.11.1999                        }
{Comment:   NC OCI8 VCL: SQL expressions, filters       }
{Copyright: Portions - (c) 1999, Nica-Com Ltd.          }
{Author:    Dmitry Arefiev, diman@ncom.ru               }
{           Based on Borland Delphi DBCommon.pas        }
{*******************************************************}
{$I NCOciDef.inc}
{*$DEFINE DEBUG}

{TODO
1. TOCIExpressionNode.FDataType currently does not used in Execute,
   but we must for Oracle like type casting
2. Current implementation use VARIANT for run time,
   it is 1.5 times slow as BDE
3. More BLOB support
4. Filter translation into server Oracle SQL
5. More error condition checking
TODO}

unit NCOciFilter;

interface

Uses Classes, DB;

type
    TOCIExpressionToken = (etEnd, etSymbol, etName, etLiteral,  etLParen,
        etRParen, etEQ, etNE, etGE, etLE, etGT, etLT, etCONCAT, etADD,
        etSUB, etMUL, etDIV, etComma, etLIKE, etISNULL, etISNOTNULL, etIN,
        etNOTIN, etANY, etSOME, etALL, etNOTLIKE, etBETWEEN, etNOTBETWEEN);

    TOCIExpressionOperator = (canNOTDEFINED, canASSIGN, canOR, canAND, canNOT,
        canEQ, canNE, canGE, canLE, canGT, canLT, canLIKE, canISBLANK, canNOTBLANK,
        canIN, canCONCAT, canADD, canSUB, canMUL, canDIV, canNOTIN, canANY, 
        canALL, canNOTLIKE, canBETWEEN, canNOTBETWEEN);

    TOCIExpressionOption = (poExtSyntax, poAggregate, poDefaultExpr,
        poFieldNameGiven);
    TOCIExpressionOptions = set of TOCIExpressionOption;

    TOCIExpressionNodeKind = (enUnknown, enField, enConst, enOperator, enFunc);
    TOCIExpressionScopeKind = (skUnknown, skField, skAgg, skConst);

    POCIExpressionNode = ^TOCIExpressionNode;
    TOCIExpressionNode = record
        FNext: POCIExpressionNode;
        FLeft: POCIExpressionNode;
        FRight: POCIExpressionNode;
        FDataType: TFieldType;
        FScopeKind: TOCIExpressionScopeKind;
        FKind: TOCIExpressionNodeKind;
        FOperator: TOCIExpressionOperator;
        FData: Variant;
        FFuncInd: Integer;
        FField: TField;
        FArgs: TList;
    end;

    TOCIExpression = class
    private
        FRoot, FNodes: POCIExpressionNode;
        FDataSet: TDataSet;
        FOptions: TFilterOptions;
        FFieldName: String;
        FParserOptions: TOCIExpressionOptions;
        function NewNode(Kind: TOCIExpressionNodeKind; Operator: TOCIExpressionOperator;
            const Data: Variant; Left, Right: POCIExpressionNode; AFuncInd: Integer = -1): POCIExpressionNode;
        procedure ClearNodes;
    public
        constructor Create(DataSet: TDataSet; Options: TFilterOptions;
            ParseOptions: TOCIExpressionOptions; const FieldName: string);
        destructor Destroy; override;
        function NewCompareNode(Field: TField; Operator: TOCIExpressionOperator;
            const Value: Variant): POCIExpressionNode;
        procedure Dump(AStrs: TStrings);
        function Execute: Variant;
    end;

    TOCIExpressionParser = class
    private
        FExpression: TOCIExpression;
        FText: string;
        FSourcePtr: PChar;
        FTokenPtr: PChar;
        FTokenString: string;
        FToken: TOCIExpressionToken;
        FNumericLit: Boolean;
        FParserOptions: TOCIExpressionOptions;
        FDataSet: TDataSet;
        procedure NextToken;
        function NextTokenIsLParen : Boolean;
        function ParseExpr: POCIExpressionNode;
        function ParseExpr2: POCIExpressionNode;
        function ParseExpr3: POCIExpressionNode;
        function ParseExpr4: POCIExpressionNode;
        function ParseExpr5(AEnableList: Boolean): POCIExpressionNode;
        function ParseExpr6: POCIExpressionNode;
        function ParseExpr7: POCIExpressionNode;
        function TokenName: string;
        function TokenSymbolIs(const S: string): Boolean;
        function TokenSymbolIsFunc(const S: string): Integer;
        procedure GetFuncResultInfo(Node: POCIExpressionNode);
        procedure TypeCheckArithOp(Node: POCIExpressionNode);
        procedure GetScopeKind(Root, Left, Right : POCIExpressionNode);
    public
        function BuildExpression(ADataSet: TDataSet; const Text: string;
            Options: TFilterOptions; ParserOptions: TOCIExpressionOptions;
            const FieldName: string): TOCIExpression;
    end;

    TOCIFilter = class(TCollectionItem)
    private
        FName: String;
        FText: String;
        FOptions: TFilterOptions;
        FActive: Boolean;
        FExpression: TOCIExpression;
        FOnFilterRecord: TFilterRecordEvent;
        FPrepared: Boolean;
        procedure SetActive(const Value: Boolean);
        procedure SetOptions(const Value: TFilterOptions);
        procedure SetText(const Value: String);
        function GetDataSet: TDataSet;
        procedure SetFilterRecord(const Value: TFilterRecordEvent);
        procedure Changed(ACanCache: Boolean);
    protected
        function GetDisplayName: string; override;
        function Execute: Boolean;
    public
        destructor Destroy; override;
        procedure Assign(AValue: TPersistent); override;
        function IsEqual(AValue: TOCIFilter): Boolean;
        procedure Prepare;
        procedure PrepareLookupFilter(const AKeyFields: String;
            const AKeyValues: Variant; AOptions: TLocateOptions);
        procedure UnPrepare;
        property DataSet: TDataSet read GetDataSet;
    published
        property Name: String read FName write FName;
        property Text: String read FText write SetText;
        property Options: TFilterOptions read FOptions write SetOptions default [];
        property Active: Boolean read FActive write SetActive default False;
        property OnFilterRecord: TFilterRecordEvent read FOnFilterRecord
            write SetFilterRecord;
    end;

    TOCIFilters = class(TCollection)
    private
        FOwner: TPersistent;
        function GetDataSet: TDataSet;
        function GetItem(AIndex: Integer): TOCIFilter;
        procedure SetItem(AIndex: Integer; AValue: TOCIFilter);
        function GetDefaultFilter: TOCIFilter;
        function GetLookupFilter: TOCIFilter;
        function GetAnyActive: Boolean;
    protected
        function GetOwner: TPersistent; override;
        procedure Update(Item: TCollectionItem); override;
    public
        constructor Create(AOwner: TPersistent); overload;
        function IsEqual(Value: TOciFilters): Boolean;
        function FilterByName(const Value: string): TOciFilter;
        function FindFilter(const Value: string): TOciFilter;
        function Execute: Boolean;
        property AnyActive: Boolean read GetAnyActive;
        property DataSet: TDataSet read GetDataSet;
        property DefaultFilter: TOCIFilter read GetDefaultFilter;
        property LookupFilter: TOCIFilter read GetLookupFilter;
        property Items[AIndex: Integer]: TOCIFilter read GetItem write SetItem; default;
        property UpdateCount;
    end;

    POCIFuncCall = function (const AArgs: Variant): Variant;
    procedure OCIAddFunc(const AName: String; AScopeKind: TOCIExpressionScopeKind;
        AScopeKindArg: Integer; ADataType: TFieldType; ADataTypeArg: Integer;
        AArgMin: Integer; AArgMax: Integer; ACall: POCIFuncCall);

implementation

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

uses Forms, SysUtils, DBConsts, NCOci, NCOciUtil, NCOciMsg, NCOciDB, Math,
     Windows;

type
    __TOCIDataSet = class(TOCIDataSet);
    TOCIFuncDesc = class
    private
        FScopeKind: TOCIExpressionScopeKind;
        FScopeKindArg: Integer;
        FDataType: TFieldType;
        FDataTypeArg: Integer;
        FArgMin: Integer;
        FArgMax: Integer;
        FCall: POCIFuncCall;
    end;

const
    FStrTrue: String = STextTrue;
    FStrFalse: String = STextFalse;
    cWordSet: set of char = ['A'..'Z', 'a'..'z'];

var
    OCIFuncs: TStringList;

function IsNumeric(DataType: TFieldType): Boolean;
begin
  Result := DataType in [ftSmallInt, ftInteger, ftFloat, ftBCD];
end;

function IsTemporal(DataType: TFieldType): Boolean;
begin
  Result := DataType in [ftDateTime];
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIFuncDesc

function StrIsNull(const V: Variant): Boolean;
var
    tp: Integer;
begin
    tp := (VarType(V) and varTypeMask);
    Result := (tp = varEmpty) or (tp = varNull) or
        ((tp = varString) or (tp = varOleStr)) and (V = '');
end;

function StrToVar(const S: String): Variant;
begin
    if S = '' then
        Result := Null
    else
        Result := S;
end;

function FunUpper(const AArgs: Variant): Variant;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else
        Result := AnsiUpperCase(AArgs[0]);
end;

function FunLower(const AArgs: Variant): Variant;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else
        Result := AnsiUpperCase(AArgs[0]);
end;

function FunSubstring(const AArgs: Variant): Variant;
var
    s: String;
    ind, cnt: Integer;
begin
    if StrIsNull(AArgs[0]) or VarIsNull(AArgs[1]) then
        Result := Null
    else begin
        s := AArgs[0];
        ind := AArgs[1];
        if ind < 0 then
            ind := Length(s) + ind + 1;
        if VarArrayHighBound(AArgs, 1) = 1 then
            cnt := Length(s)
        else if VarIsNull(AArgs[2]) or (AArgs[2] <= 0) then begin
            Result := Null;
            Exit;
        end
        else
            cnt := AArgs[2];
        Result := StrToVar(Copy(s, ind, cnt));
    end;
end;

type
    TOCITrimMode = set of (tmLeft, tmRight);

function InternalTrim(const AArgs: Variant; AMode: TOCITrimMode): Variant;
var
    I, L: Integer;
    sWhere, sWhat: String;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else begin
        sWhere := AArgs[0];
        if VarArrayHighBound(AArgs, 1) = 1 then begin
            if StrIsNull(AArgs[1]) then begin
                Result := Null;
                Exit;
            end
            else
                sWhat := AArgs[1];
        end
        else
            sWhat := ' ';
        L := Length(sWhere);
        I := 1;
        if tmLeft in AMode then
            while (I <= L) and (StrScan(PChar(sWhat), sWhere[I]) <> nil) do
                Inc(I);
        if I > L then
            sWhere := ''
        else begin
            if tmRight in AMode then
                while (L >= I) and (StrScan(PChar(sWhat), sWhere[L]) <> nil) do
                    Dec(L);
            sWhere := Copy(sWhere, I, L - I + 1);
        end;
        Result := StrToVar(sWhere);
    end;
end;

function FunTrim(const AArgs: Variant): Variant;
begin
    Result := InternalTrim(AArgs, [tmLeft, tmRight]);
end;

function FunTrimLeft(const AArgs: Variant): Variant;
begin
    Result := InternalTrim(AArgs, [tmLeft]);
end;

function FunTrimRight(const AArgs: Variant): Variant;
begin
    Result := InternalTrim(AArgs, [tmRight]);
end;

function FunYear(const AArgs: Variant): Variant;
var
    Y, M, D: Word;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeDate(AArgs[0], Y, M, D);
        Result := Y;
    end;
end;

function FunMonth(const AArgs: Variant): Variant;
var
    Y, M, D: Word;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeDate(AArgs[0], Y, M, D);
        Result := M;
    end;
end;

function FunDay(const AArgs: Variant): Variant;
var
    Y, M, D: Word;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeDate(AArgs[0], Y, M, D);
        Result := D;
    end;
end;

function FunHour(const AArgs: Variant): Variant;
var
    H, M, S, MS: Word;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeTime(AArgs[0], H, M, S, MS);
        Result := H;
    end;
end;

function FunMinute(const AArgs: Variant): Variant;
var
    H, M, S, MS: Word;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeTime(AArgs[0], H, M, S, MS);
        Result := M;
    end;
end;

function FunSecond(const AArgs: Variant): Variant;
var
    H, M, S, MS: Word;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeTime(AArgs[0], H, M, S, MS);
        Result := S;
    end;
end;

function FunGetDate(const AArgs: Variant): Variant;
begin
    Result := Date;
end;

function FunDate(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else
        Result := Integer(Trunc(AArgs[0]));
end;

function FunTime(const AArgs: Variant): Variant;
var
    dt: TDateTime;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        dt := AArgs[0];
        Result := dt - Trunc(dt);
    end;
end;

function FunAbs(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Abs(AArgs[0]);
    end;
end;

function FunCeil(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Ceil(AArgs[0]);
    end;
end;

function FunCos(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Cos(AArgs[0]);
    end;
end;

function FunCosh(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Cosh(AArgs[0]);
    end;
end;

function FunExp(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Exp(AArgs[0]);
    end;
end;

function FunFloor(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Floor(AArgs[0]);
    end;
end;

function FunLn(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := Ln(AArgs[0]);
    end;
end;

function FunLog(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) or VarIsNull(AArgs[1]) then
        Result := Null
    else begin
        Result := LogN(AArgs[0], AArgs[1]);
    end;
end;

function FunMod(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) or VarIsNull(AArgs[1]) then
        Result := Null
    else begin
        Result := AArgs[0] mod AArgs[1]
    end;
end;

function FunPower(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) or VarIsNull(AArgs[1]) then
        Result := Null
    else begin
        Result := Power(AArgs[0], AArgs[1]);
    end;
end;

function FunRound(const AArgs: Variant): Variant;
begin
{TODO
    Support for
    1) for number
    2) for date
    3) Arg#2 - format mask
}
    if VarIsNull(AArgs[0]) then
        Result := Null
    else
        Result := Round(AArgs[0]) + 0.0;
end;

function FunSign(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        if AArgs[0] > 0.0 then
            Result := 1
        else if AArgs[0] < 0.0 then
            Result := -1
        else
            Result := 0;
    end;
end;

function FunSin(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := sin(AArgs[0]);
    end;
end;

function FunSinh(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := sinh(AArgs[0]);
    end;
end;

function FunSqrt(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := sqrt(AArgs[0]);
    end;
end;

function FunTan(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := tan(AArgs[0]);
    end;
end;

function FunTanh(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        Result := tanh(AArgs[0]);
    end;
end;

function FunTrunc(const AArgs: Variant): Variant;
begin
{TODO
    Support for
    1) for number
    2) for date
    3) Arg#2 - format mask
}
    if VarIsNull(AArgs[0]) then
        Result := Null
    else
        Result := Trunc(AArgs[0]) + 0.0;
end;

function FunChr(const AArgs: Variant): Variant;
var
    s: String;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        s := Chr(Integer(AArgs[0]));
        Result := s;
    end;
end;

function FunConcat(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := AArgs[1]
    else if VarIsNull(AArgs[1]) then
        Result := AArgs[0]
    else begin
        Result := Concat(AArgs[0], AArgs[1]);
    end;
end;

function FunInitCap(const AArgs: Variant): Variant;
var
    i: Integer;
    s: String;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else begin
        s := AArgs[0];
        for i := 0 to Length(s) do
            if (s[i] in cWordSet) and ((i = 1) or not (s[i - 1] in cWordSet)) then
                s[i] := Chr(Integer(CharUpper(PChar(Ord(s[i])))));
    end;
end;

function InternalPad(const AArgs: Variant; AFront: Boolean): Variant;
var
    n: Integer;
    s, ps: String;
begin
    if StrIsNull(AArgs[0]) or VarIsNull(AArgs[1]) or            
       (VarArrayHighBound(AArgs, 1) = 2) and StrIsNull(AArgs[2]) then
        Result := Null
    else begin
        s := AArgs[0];
        n := AArgs[1] - Length(s);
        if (VarArrayHighBound(AArgs, 1) = 1) then
            ps := ' '
        else
            ps := AArgs[2];
        while Length(ps) < n do
            ps := ps + ps;
        if Length(ps) > n then
            ps := Copy(ps, 1, n);
        if AFront then
            Result := StrToVar(ps + s)
        else
            Result := StrToVar(s + ps);
    end;
end;

function FunLPad(const AArgs: Variant): Variant;
begin
    Result := InternalPad(AArgs, True);
end;

function FunRPad(const AArgs: Variant): Variant;
begin
    Result := InternalPad(AArgs, False);
end;

function FunReplace(const AArgs: Variant): Variant;
var
    sWhere, sFrom, sTo: String;
    i: Integer;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else if StrIsNull(AArgs[1]) then
        Result := AArgs[0]
    else begin
        sWhere := AArgs[0];
        sFrom := AArgs[1];
        if VarArrayHighBound(AArgs, 1) = 2 then
            sTo := AArgs[2]
        else
            sTo := '';
        while True do begin
            i := Pos(sFrom, sWhere);
            if i = 0 then
                Break;
            Delete(sWhere, i, Length(sFrom));
            if sTo <> '' then
                Insert(sTo, sWhere, i);
        end;
        Result := StrToVar(sWhere);
    end;
end;

function FunSoundex(const AArgs: Variant): Variant;
const
    SOUNDEX_LENGTH = 4;
var
    Source: String;
    code: String[SOUNDEX_LENGTH];
    i, j: Integer;
    c, res, prev: Char;

    function step2(C: Char): Char;
    begin
        case c of
        'B', 'F', 'P', 'V':
            Result := '1';
        'C', 'G', 'J', 'K', 'Q', 'S', 'X', 'Z':
            Result := '2';
        'D', 'T':
            Result := '3';
        'L':
            Result := '4';
        'M', 'N':
            Result := '5';
        'R':
            Result := '6';
        'A', 'E', 'H', 'I', 'O', 'U', 'W', 'Y':
            Result := #0;
        else
            Result := '9';
        end;
    end;

begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else begin
        Source := Trim(AArgs[0]);
        if Source = '' then
            Result := Null
        else begin
            prev := #0;
            i := 1;
            j := 1;
            while (i <= Length(Source)) and (j <= SOUNDEX_LENGTH) do begin
                c := Chr(Integer(CharUpper(PChar(Ord(source[i])))));
                if c in cWordSet then begin
                    res := Step2(c);
                    if j = 1 then begin
                        code[j] := c;
                        Inc(j);
                    end
                    else if (res <> #0) and (res <> prev) then begin
                        code[j] := res;
                        Inc(j);
                    end;
                    prev := res;
                end;
                Inc(i);
            end;
            while j <= SOUNDEX_LENGTH do begin
                code[j] := '0';
                Inc(j);
            end;
            Result := code;
        end;
    end;
end;

function FunTranslate(const AArgs: Variant): Variant;
var
    sWhere, sFrom, sTo: String;
    i, j: Integer;
    pCh: PChar;
begin
    if StrIsNull(AArgs[0]) or StrIsNull(AArgs[1]) or StrIsNull(AArgs[2]) then
        Result := Null
    else begin
        sWhere := AArgs[0];
        sFrom := AArgs[1];
        sTo := AArgs[2];
        i := 1;
        while i <= Length(sWhere) do begin
            pCh := StrScan(PChar(sFrom), sWhere[i]);
            if pCh <> nil then begin
                j := pCh - PChar(sFrom) + 1;
                if j > Length(sTo) then
                    Delete(sWhere, i, 1)
                else
                    sWhere[i] := sTo[j];
            end;
        end;
        Result := StrToVar(sWhere);
    end;
end;

function FunAscii(const AArgs: Variant): Variant;
var
    s: String;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else begin
        s := AArgs[0];
        Result := Ord(s[1]);
    end;
end;

function FunInstr(const AArgs: Variant): Variant;
var
    sWhere, sWhat: String;
    iCount, iFrom, hBnd: Integer;
    pStart, pCh: PChar;
    v: Variant;
    bReverse: Boolean;
begin
    if StrIsNull(AArgs[0]) or StrIsNull(AArgs[1]) then
        Result := Null
    else begin
        sWhere := AArgs[0];
        sWhat := AArgs[1];
        hBnd := VarArrayHighBound(AArgs, 1);
        iCount := 1;
        iFrom := 1;
        bReverse := False;
        if hBnd >= 2 then begin
            v := AArgs[2];
            if VarIsNull(v) then begin
                Result := Null;
                Exit;
            end;
            iFrom := v;
            bReverse := iFrom < 0;
            if bReverse then
                iFrom := Length(sWhere) + iFrom + 1;
            if hBnd >= 3 then begin
                v := AArgs[3];
                if VarIsNull(v) or (v <= 0) then begin
                    Result := Null;
                    Exit;
                end;
                iCount := v;
            end;
        end;
        pStart := PChar(sWhere) + iFrom - 1;
        pCh := nil;
        while iCount > 0 do begin
            if bReverse then begin
                pCh := StrRPos(pStart, PChar(sWhat));
                if pCh = nil then
                    Break;
                pStart := pCh - Length(sWhat);
            end
            else begin
                pCh := StrPos(pStart, PChar(sWhat));
                if pCh = nil then
                    Break;
                pStart := pCh + Length(sWhat);
            end;
            Dec(iCount);
        end;
        if pCh = nil then
            Result := 0
        else
            Result := pCh - PChar(sWhere) + 1;
    end;
end;

function FunLength(const AArgs: Variant): Variant;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else
        Result := Length(AArgs[0]);
end;

function FunAddMonths(const AArgs: Variant): Variant;
var
    AYear, AMonth, ADay: Word;
begin
    if StrIsNull(AArgs[0]) or StrIsNull(AArgs[1]) then
        Result := Null
    else begin
        DecodeDate(AArgs[0], AYear, AMonth, ADay);
        AMonth := (AYear + (AYear - 1)) * 12 + (AMonth + AArgs[1] - 1);
        AYear := AMonth div 12 + 1;
        AMonth := AMonth mod 12 + 1;
        Result := EncodeDate(AYear, AMonth, ADay);
    end;
end;

function FunLastDay(const AArgs: Variant): Variant;
const
    DaysPerMonth: array[1..12] of Integer =
        (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
var
    AYear, AMonth, ADay: Word;
    t: TDateTime;
begin
    if StrIsNull(AArgs[0]) then
        Result := Null
    else begin
        DecodeDate(AArgs[0], AYear, AMonth, ADay);
        t := AArgs[0];
        ADay := DaysPerMonth[AMonth];
        if (AMonth = 2) and
           (AYear mod 4 = 0) and ((AYear mod 100 <> 0) or (AYear mod 400 = 0)) then
            Inc(ADay);
        Result := EncodeDate(AYear, AMonth, ADay) + t;
    end;
end;

function FunMonthsBW(const AArgs: Variant): Variant;
var
    AYear1, AMonth1, ADay1: Word;
    AYear2, AMonth2, ADay2: Word;
begin
    if StrIsNull(AArgs[0]) or StrIsNull(AArgs[1]) then
        Result := Null
    else begin
        DecodeDate(AArgs[0], AYear1, AMonth1, ADay1);
        DecodeDate(AArgs[1], AYear2, AMonth2, ADay2);
        Result := (AYear1 * 12 + AMonth1) - (AYear2 * 12 + AMonth2);
    end;
end;

function FunToChar(const AArgs: Variant): Variant;
var
    s: String;
begin
{TODO support of mask}
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        s := AArgs[0];
        Result := s;
    end;
end;

function FunToDate(const AArgs: Variant): Variant;
var
    d: TDateTime;
begin
{TODO support of mask}
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        d := AArgs[0];
        Result := d;
    end;
end;

function FunToNumber(const AArgs: Variant): Variant;
var
    f: Double;
begin
{TODO support of mask}
    if VarIsNull(AArgs[0]) then
        Result := Null
    else begin
        f := AArgs[0];
        Result := f;
    end;
end;

function FunDecode(const AArgs: Variant): Variant;
var
    n, i: Integer;
begin
    n := VarArrayHighBound(AArgs, 1);
    i := 1;
    Result := AArgs[0];
    while i <= n - 1 do begin
        if Result = AArgs[i] then begin
            Result := AArgs[i + 1];
            Break;
        end;
        Inc(i, 2);
    end;
    if i = n then
        Result := AArgs[i];
end;

function FunIIF(const AArgs: Variant): Variant;
begin
    if VarIsNull(AArgs[0]) then
        Result := Null
    else if AArgs[0] then
        Result := AArgs[1]
    else
        Result := AArgs[2]
end;

function FunNvl(const AArgs: Variant): Variant;
begin
    Result := AArgs[0];
    if VarIsNull(Result) then
        Result := AArgs[1];
end;

function FunGreatest(const AArgs: Variant): Variant;
var
    n, i: Integer;
begin
    n := VarArrayHighBound(AArgs, 1);
    Result := AArgs[0];
    for i := 0 to n do begin
        if VarIsNull(AArgs[i]) then begin
            Result := Null;
            Break;
        end;
        if Result < AArgs[i] then
            Result := AArgs[i];
    end;
end;

function FunLeast(const AArgs: Variant): Variant;
var
    n, i: Integer;
begin
    n := VarArrayHighBound(AArgs, 1);
    Result := AArgs[0];
    for i := 0 to n do begin
        if VarIsNull(AArgs[i]) then begin
            Result := Null;
            Break;
        end;
        if Result > AArgs[i] then
            Result := AArgs[i];
    end;
end;

procedure OCIAddFunc(const AName: String; AScopeKind: TOCIExpressionScopeKind;
    AScopeKindArg: Integer; ADataType: TFieldType; ADataTypeArg: Integer;
    AArgMin: Integer; AArgMax: Integer; ACall: POCIFuncCall);
var
    pFuncDesc: TOCIFuncDesc;
begin
    pFuncDesc := TOCIFuncDesc.Create;
    with pFuncDesc do begin
        FScopeKind := AScopeKind;
        FScopeKindArg := AScopeKindArg;
        FDataType := ADataType;
        FDataTypeArg := ADataTypeArg;
        FArgMin := AArgMin;
        FArgMax := AArgMax;
        FCall := ACall;
    end;
    OCIFuncs.AddObject(UpperCase(AName), pFuncDesc);
end;

function GetFuncDesc(AIndex: Integer): TOCIFuncDesc;
begin
    Result := TOCIFuncDesc(OCIFuncs.Objects[AIndex]);
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIExpression

constructor TOCIExpression.Create(DataSet: TDataSet; Options: TFilterOptions;
    ParseOptions: TOCIExpressionOptions; const FieldName: string);
begin
    inherited Create;
    FDataSet := DataSet;
    FOptions := Options;
    FFieldName := FieldName;
    FParserOptions := ParseOptions;
end;

destructor TOCIExpression.Destroy;
begin
    ClearNodes;
    inherited Destroy;
end;

function TOCIExpression.NewNode(Kind: TOCIExpressionNodeKind; Operator: TOCIExpressionOperator;
    const Data: Variant; Left, Right: POCIExpressionNode; AFuncInd: Integer = -1): POCIExpressionNode;
begin
    New(Result);
    with Result^ do begin
        FNext := FNodes;
        FKind := Kind;
        FOperator := Operator;
        FData := Data;
        FLeft := Left;
        FRight := Right;
        FArgs := nil;
        FDataType := ftUnknown;
        FScopeKind := skUnknown;
        FFuncInd := AFuncInd;
        FField := nil;
    end;
    FNodes := Result;
    if Kind = enField then begin
        if poFieldNameGiven in FParserOptions then
            Result^.FField := FDataSet.FieldByName(FFieldName);
        if Result^.FField = nil then
            Result^.FField := FDataSet.FieldByName(Data);
        if Result^.FField = nil then
            DatabaseErrorFmt(SFieldNotFound, [Data]);
        Result^.FDataType := Result^.FField.DataType;
    end;
end;

function TOCIExpression.NewCompareNode(Field: TField; Operator: TOCIExpressionOperator;
    const Value: Variant): POCIExpressionNode;
var
    ConstExpr: POCIExpressionNode;
begin
    ConstExpr := NewNode(enConst, canNOTDEFINED, Value, nil, nil);
    ConstExpr^.FDataType := Field.DataType;
    Result := NewNode(enOperator, Operator, Unassigned,
        NewNode(enField, canNOTDEFINED, Field.FieldName, nil, nil),
        ConstExpr);
end;

procedure TOCIExpression.ClearNodes;
var
    Node: POCIExpressionNode;
begin
    while FNodes <> nil do begin
        Node := FNodes;
        FNodes := Node^.FNext;
        if Node^.FArgs <> nil then
            Node^.FArgs.Free;
        Dispose(Node);
    end;
end;

function TOCIExpression.Execute: Variant;

    function EvaluteNodeValueComplex(ANode: POCIExpressionNode): Variant;
    var
        l, r: Variant;
        i: Integer;
        pR: POCIExpressionNode;
        esc: String;
        pFunc: TOCIFuncDesc;

        function VarIsString(const V: Variant): Boolean;
        var
            tp: Integer;
        begin
            tp := VarType(l);
            Result := (tp = varString) or (tp = varOleStr);
        end;

        procedure UpperCaseLR;
        begin
            if VarIsString(l) then
                l := AnsiUpperCase(l);
            if VarIsString(r) then
                r := AnsiUpperCase(r);
        end;

        function PartialEQ: Boolean;
        var
            sL, sR: String;
            ln, lnL, lnR: Integer;
            partial: Boolean;
        begin
            if VarIsString(l) and VarIsString(r) then begin
                sL := l;
                sR := r;
                lnL := Length(sL);
                lnR := Length(sR);
                if l <> '' then begin
                    partial := False;
                    if sL[lnL] = '*' then begin
                        partial := True;
                        Dec(lnL);
                    end;
                    if r <> '' then begin
                        if sR[lnR] = '*' then begin
                            partial := True;
                            Dec(lnR);
                        end;
                        if partial then begin
                            ln := lnR;
                            if ln > lnL then
                                ln := lnL;
                            if (foCaseInsensitive in FOptions) then
                                Result := AnsiStrLIComp(PChar(sL), PChar(sR), ln) = 0
                            else
                                Result := AnsiStrLComp(PChar(sL), PChar(sR), ln) = 0;
                            Exit;
                        end;
                    end;
                end;
                if (foCaseInsensitive in FOptions) then
                    Result := AnsiCompareText(sL, sR) = 0
                else
                    Result := sL = Sr;
            end
            else begin
                UpperCaseLR;
                Result := l = r;
            end;
        end;

        function EvaluteNodeValue(ANode: POCIExpressionNode): Variant;
        begin
            if ANode^.FKind = enField then
                Result := ANode^.FField.Value
            else if ANode^.FKind = enConst then
                Result := ANode^.FData
            else
                Result := EvaluteNodeValueComplex(ANode);
        end;

    begin
        case ANode^.FKind of
        enUnknown:;
        enField:
            Result := ANode^.FField.Value;
        enConst:
            Result := ANode^.FData;
        enOperator:
            case ANode^.FOperator of
            canNOTDEFINED:;
            canASSIGN:
                FDataSet.FieldByName(ANode^.FRight^.FData).Value :=
                    EvaluteNodeValue(ANode^.FLeft);
            canOR:
                Result := Boolean(EvaluteNodeValue(ANode^.FLeft)) or
                          Boolean(EvaluteNodeValue(ANode^.FRight));
            canAND:
                Result := Boolean(EvaluteNodeValue(ANode^.FLeft)) and
                          Boolean(EvaluteNodeValue(ANode^.FRight));
            canNOT:
                Result := not Boolean(EvaluteNodeValue(ANode^.FLeft));
            canEQ, canNE, canGE, canLE, canGT, canLT:
                begin
                    l := EvaluteNodeValue(ANode^.FLeft);
                    pR := ANode^.FRight;
                    if pR^.FOperator in [canANY, canALL] then begin
                        for i := 0 to pR.FArgs.Count - 1 do begin
                            r := EvaluteNodeValue(POCIExpressionNode(pR.FArgs[i]));
                            if (foCaseInsensitive in FOptions) and
                               not (foNoPartialCompare in FOptions) then
                                UpperCaseLR;
                            case ANode^.FOperator of
                                canEQ:
                                    if foNoPartialCompare in FOptions then
                                        Result := l = r
                                    else
                                        Result := PartialEQ;
                                canNE: Result := l <> r;
                                canGE: Result := l >= r;
                                canLE: Result := l <= r;
                                canGT: Result := l > r;
                                canLT: Result := l < r;
                            end;
                            if (pR^.FOperator = canANY) and Result or
                               (pR^.FOperator = canALL) and not Result then
                                Break;
                        end;
                    end
                    else begin
                        r := EvaluteNodeValue(ANode^.FRight);
                        if (foCaseInsensitive in FOptions) and
                           not (foNoPartialCompare in FOptions) then
                            UpperCaseLR;
                        case ANode^.FOperator of
                            canEQ:
                                if foNoPartialCompare in FOptions then
                                    Result := l = r
                                else
                                    Result := PartialEQ;
                            canNE: Result := l <> r;
                            canGE: Result := l >= r;
                            canLE: Result := l <= r;
                            canGT: Result := l > r;
                            canLT: Result := l < r;
                        end;
                    end;
                end;
            canLIKE, canNOTLIKE:
                begin
                    if ANode^.FArgs <> nil then
                        esc := EvaluteNodeValue(POCIExpressionNode(ANode^.FArgs.Items[0]))
                    else
                        esc := #9;
                    Result := StrLike(EvaluteNodeValue(ANode^.FLeft), EvaluteNodeValue(ANode^.FRight),
                                      foCaseInsensitive in FOptions, '%', '_', esc[1]);
                    if ANode^.FOperator = canNOTLIKE then
                        Result := not Result;
                end;
            canISBLANK, canNOTBLANK:
                begin
                    Result := VarIsNull(EvaluteNodeValue(ANode^.FLeft));
                    if ANode^.FOperator = canNOTBLANK then
                        Result := not Result;
                end;
            canIN, canNOTIN:
                begin
                    Result := False;
                    l := EvaluteNodeValue(ANode^.FLeft);
                    for i := 0 to ANode^.FArgs.Count - 1 do
                        if l = EvaluteNodeValue(POCIExpressionNode(ANode^.FArgs.Items[i])) then begin
                            Result := True;
                            Break;
                        end;
                    if ANode^.FOperator = canNOTIN then
                        Result := not Result;
                end;
            canCONCAT, canADD:
                Result := EvaluteNodeValue(ANode^.FLeft) +
                          EvaluteNodeValue(ANode^.FRight);
            canSUB:
                Result := EvaluteNodeValue(ANode^.FLeft) -
                          EvaluteNodeValue(ANode^.FRight);
            canMUL:
                Result := EvaluteNodeValue(ANode^.FLeft) *
                          EvaluteNodeValue(ANode^.FRight);
            canDIV:
                Result := EvaluteNodeValue(ANode^.FLeft) /
                          EvaluteNodeValue(ANode^.FRight);
            canBETWEEN, canNOTBETWEEN:
                begin
                    l := EvaluteNodeValue(ANode^.FLeft);
                    Result := (l >= EvaluteNodeValue(POCIExpressionNode(ANode^.FArgs.Items[0]))) and
                              (l <= EvaluteNodeValue(POCIExpressionNode(ANode^.FArgs.Items[1])));
                    if ANode^.FOperator = canNOTBETWEEN then
                        Result := not Result;
                end;
            end;
        enFunc:
            begin
                pFunc := GetFuncDesc(ANode^.FFuncInd);
                if pFunc.FScopeKind = skAgg then begin
                    raise Exception.Create('Aggregate functions not ready ;-((');
{TODO               if pFunc.FName = 'SUM' then
                    else if pFunc.FName = 'MIN' then
                    else if pFunc.FName = 'MAX' then
                    else if pFunc.FName = 'AVG' then
                    else if pFunc.FName = 'COUNT' then
                    ;
}               end
                else begin
                    if (ANode^.FArgs = nil) or (ANode^.FArgs.Count = 0) then
                        r := null
                    else begin
                        r := VarArrayCreate([0, ANode^.FArgs.Count - 1], varVariant);
                        for i := 0 to ANode^.FArgs.Count - 1 do
                            r[i] := EvaluteNodeValue(POCIExpressionNode(ANode^.FArgs.Items[i]));
                    end;
                    Result := pFunc.FCall(r);
                end;
            end;
        end;
    end;
begin
    Result := EvaluteNodeValueComplex(FRoot);
end;

procedure TOCIExpression.Dump(AStrs: TStrings);
{$IFDEF DEBUG}
const
    enk2s: array[TOCIExpressionNodeKind] of String = (
        'enUnknown', 'enField', 'enConst', 'enOperator', 'enFunc');
    bool2s: array[Boolean] of String = (
        'False', 'True');
    esk2s: array[TOCIExpressionScopeKind] of String = (
        'skUnknown', 'skField', 'skAgg', 'skConst'
    );
    co2s: array[TOCIExpressionOperator] of String = (
        'canNOTDEFINED', 'canASSIGN', 'canOR', 'canAND', 'canNOT', 'canEQ', 'canNE',
        'canGE', 'canLE', 'canGT', 'canLT', 'canLIKE', 'canISBLANK', 'canNOTBLANK',
        'canIN', 'canCONCAT', 'canADD', 'canSUB', 'canMUL', 'canDIV', 'canNOTIN',
        'canANY', 'canALL', 'canNOTLIKE', 'canBETWEEN', 'canNOTBETWEEN'
    );
    ft2s: array[TFieldType] of String = (
        'ftUnknown', 'ftString', 'ftSmallint', 'ftInteger', 'ftWord', 'ftBoolean',
        'ftFloat', 'ftCurrency', 'ftBCD', 'ftDate', 'ftTime', 'ftDateTime', 'ftBytes',
        'ftVarBytes', 'ftAutoInc', 'ftBlob', 'ftMemo', 'ftGraphic', 'ftFmtMemo',
        'ftParadoxOle', 'ftDBaseOle', 'ftTypedBinary', 'ftCursor', 'ftFixedChar',
        'ftWideString', 'ftLargeint', 'ftADT', 'ftArray', 'ftReference', 'ftDataSet'
{$IFDEF OCI_D5}
        , 'ftOraBlob', 'ftOraClob', 'ftVariant', 'ftInterface', 'ftIDispatch', 'ftGuid'
{$ENDIF}
        );

    function Var2Text(const AValue: Variant): String;
    begin
        if VarIsEmpty(AValue) then
            Result := 'Unassigned'
        else if VarIsNull(AValue) then
            Result := 'Null'
        else
            Result := AValue;
    end;

    procedure DumpNode(ANode: POCIExpressionNode; AIndent: Integer; AList: TStrings);
    var
        st: String;
        i: Integer;
    begin
        with ANode^ do begin
            st := '';
            for i := 1 to AIndent do
                st := st + ' ';
            AList.Add(st + 
                'Kind: ' + enk2s[FKind] + ' ' +
                'Oper: ' + co2s[Foperator] + ' ' +
                'Data: ' + Var2Text(FData) + ' ' +
                'DT: ' + ft2s[FDataType] + ' ' +
                'Scope: ' + esk2s[FScopeKind]
            );
            if (FArgs <> nil) and (FArgs.Count > 0) then begin
                AList.Add(st + 'ARGS:');
                for i := 0 to FArgs.Count - 1 do
                    DumpNode(POCIExpressionNode(FArgs[i]), AIndent + 2, AList);
            end;
            if FLeft <> nil then begin
                AList.Add(st + 'LEFT:');
                DumpNode(FLeft, AIndent + 2, AList);
            end;
            if FRight <> nil then begin
                AList.Add(st + 'RIGHT:');
                DumpNode(FRight, AIndent + 2, AList);
            end;
        end;
    end;

{$ENDIF}
begin
{$IFDEF DEBUG}
    AStrs.Clear;
    if FRoot <> nil then
        DumpNode(FRoot, 0, AStrs)
    else
        AStrs.Add('-= EMPTY =-');
{$ELSE}
    raise Exception.Create('Before use of TOCIExpression.Dump recompile NCOciFilter with $DEFINE DEBUG');
{$ENDIF}
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIExpressionParser

function TOCIExpressionParser.BuildExpression(ADataSet: TDataSet; const Text: string;
    Options: TFilterOptions; ParserOptions: TOCIExpressionOptions; const FieldName: string): TOCIExpression;
var
    Root, DefField: POCIExpressionNode;
begin
    FExpression := TOCIExpression.Create(ADataSet, Options, ParserOptions, FieldName);
    FDataSet := ADataSet;
    FParserOptions := ParserOptions;
    FText := Text;
    FSourcePtr := PChar(Text);
    NextToken;
    Root := ParseExpr;
    if FToken <> etEnd then
        DatabaseError(SExprTermination);
    if (poAggregate in FParserOptions) and (Root^.FScopeKind <> skAgg) then
        DatabaseError(SExprNotAgg);
    if (not (poAggregate in FParserOptions)) and (Root^.FScopeKind = skAgg) then
        DatabaseError(SExprNoAggFilter);
    if poDefaultExpr in ParserOptions then begin
        DefField := FExpression.NewNode(enField, canNOTDEFINED, FieldName, nil, nil);
        if (IsTemporal(DefField^.FDataType) and (Root^.FDataType = ftString)) or
           ((DefField^.FDataType = ftBoolean) and (Root^.FDataType = ftString)) then
            Root^.FDataType := DefField^.FDataType;
        if not ((IsTemporal(DefField^.FDataType) and IsTemporal(Root^.FDataType))
           or (IsNumeric(DefField^.FDataType) and IsNumeric(Root^.FDataType))
           or ((DefField^.FDataType = ftString) and (Root^.FDataType = ftString))
           or ((DefField^.FDataType = ftBoolean) and (Root^.FDataType = ftBoolean))) then
            DatabaseError(SExprTypeMis);
        Root := FExpression.NewNode(enOperator, canASSIGN, Unassigned, Root, DefField);
    end;
    if not (poAggregate in FParserOptions) and not (poDefaultExpr in ParserOptions)
       and (Root^.FDataType <> ftBoolean) then
        DatabaseError(SExprIncorrect);
    FExpression.FRoot := Root;
    Result := FExpression;
end;

function TOCIExpressionParser.NextTokenIsLParen : Boolean;
var
    P : PChar;
begin
    P := FSourcePtr;
    while (P^ <> #0) and (P^ <= ' ') do
        Inc(P);
    Result := P^ = '(';
end;

procedure TOCIExpressionParser.NextToken;
type
    ASet = Set of Char;
var
    tmpPtr, P, TokenStart: PChar;
    L: Integer;
    StrBuf: array[0..255] of Char;
    PrevToken: TOCIExpressionToken;
    ch2: Char;

    procedure Skip(TheSet: ASet);
    begin
        while (P^ <> #0) and (P^ in TheSet) do
            Inc(P);
    end;

    procedure SkipWS;
    begin
        while (P^ <> #0) and (P^ <= ' ') do
            Inc(P);
    end;

    procedure SkpiWhileNot(TheSet: ASet);
    begin
        while (P^ <> #0) and not (P^ in TheSet) do
            Inc(P);
    end;

    procedure ExtractToken(TheSet: ASet);
    begin
        SkipWS;
        TokenStart := P;
        Skip(TheSet);
        SetString(FTokenString, TokenStart, P - TokenStart);
    end;

begin
    PrevToken := FToken;
    FTokenString := '';
    P := FSourcePtr;
    SkipWS;
    // /* comment */
    if (P^ <> #0) and (P^ = '/') and (P[1] <> #0) and (P[1] = '*') then begin
        Inc(P, 2);
        SkpiWhileNot(['*']);
        if (P^ = '*') and (P[1] <> #0) and (P[1] =  '/')  then
            Inc(P, 2)
        else
            DatabaseErrorFmt(SExprInvalidChar, [P^]);
    end
    // -- comment
    else if (P^ <> #0) and (P^ = '-') and (P[1] <> #0) and (P[1] = '-') then begin
        Inc(P, 2);
        SkpiWhileNot([#13, #10]);
    end;
    SkipWS;
    FTokenPtr := P;

    case P^ of
    'A'..'Z', 'a'..'z':
        begin
            ExtractToken(['A'..'Z', 'a'..'z', '0'..'9', '$', '#', '_', '.']);
            FToken := etSymbol;
            if CompareText(FTokenString, 'NOT') = 0 then begin
                tmpPtr := P;
                ExtractToken(['A'..'Z', 'a'..'z']);
                if CompareText(FTokenString, 'IN') = 0 then
                    FToken := etNOTIN
                else if CompareText(FTokenString, 'LIKE') = 0 then
                    FToken := etNOTLIKE
                else if CompareText(FTokenString, 'BETWEEN') = 0 then
                    FToken := etNOTBETWEEN
                else
                    P := tmpPtr;
            end
            else if CompareText(FTokenString, 'LIKE') = 0 then
                FToken := etLIKE
            else if CompareText(FTokenString, 'IN') = 0 then
                FToken := etIN
            else if CompareText(FTokenString, 'ANY') = 0 then
                FToken := etANY
            else if CompareText(FTokenString, 'SOME') = 0 then
                FToken := etSOME
            else if CompareText(FTokenString, 'ALL') = 0 then
                FToken := etALL
            else if CompareText(FTokenString, 'BETWEEN') = 0 then
                FToken := etBETWEEN
            else if CompareText(FTokenString, 'IS') = 0 then begin
                ExtractToken(['A'..'Z', 'a'..'z']);
                if CompareText(FTokenString, 'NOT') = 0 then begin
                    ExtractToken(['A'..'Z', 'a'..'z']);
                    if CompareText(FTokenString, 'NULL') = 0 then
                        FToken := etISNOTNULL
                    else
                        DatabaseError(SInvalidKeywordUse);
                end
                else if CompareText (FTokenString, 'NULL') = 0  then begin
                    FToken := etISNULL;
                end
                else
                    DatabaseError(SInvalidKeywordUse);
            end;
        end;
    '[', '"':
        begin
            if P^ = '[' then
                ch2 := ']'
            else
                ch2 := '"';
            Inc(P);
            TokenStart := P;
            P := AnsiStrScan(P, ch2);
            if P = nil then
                DatabaseError(SExprNameError);
            SetString(FTokenString, TokenStart, P - TokenStart);
            FToken := etName;
            Inc(P);
        end;
    '''':
        begin
            Inc(P);
            L := 0;
            while True do begin
                if P^ = #0 then
                    DatabaseError(SExprStringError);
                if P^ = '''' then begin
                    Inc(P);
                    if P^ <> '''' then
                        Break;
                end;
                if L < SizeOf(StrBuf) then begin
                    StrBuf[L] := P^;
                    Inc(L);
                end;
                Inc(P);
            end;
            SetString(FTokenString, StrBuf, L);
            FToken := etLiteral;
            FNumericLit := False;
        end;
    '-', '0'..'9':
        begin
            if (PrevToken <> etLiteral) and (PrevToken <> etName) and
               (PrevToken <> etSymbol) and (PrevToken <> etRParen) then begin
                ExtractToken(['0'..'9', '.', 'e', 'E', '+', '-']);
                FToken := etLiteral;
                FNumericLit := True;
            end
            else begin
                FToken := etSUB;
                Inc(P);
            end;
        end;
    '(':
        begin
            Inc(P);
            FToken := etLParen;
        end;
    ')':
        begin
            Inc(P);
            FToken := etRParen;
        end;
    '<':
        begin
            Inc(P);
            case P^ of
            '=':
                begin
                    Inc(P);
                    FToken := etLE;
                end;
            '>':
                begin
                    Inc(P);
                    FToken := etNE;
                end;
            else
                FToken := etLT;
            end;
        end;
    '^', '!':
        begin
            Inc(P);
            if P^ = '=' then begin
                Inc(P);
                FToken := etNE;
            end
            else
                DatabaseErrorFmt(SExprInvalidChar, [P^]);
        end;
    '=':
        begin
            Inc(P);
            FToken := etEQ;
        end;
    '>':
        begin
            Inc(P);
            if P^ = '=' then begin
                Inc(P);
                FToken := etGE;
            end
            else
                FToken := etGT;
        end;
    '+':
        begin
            Inc(P);
            FToken := etADD;
        end;
    '|':
        begin
            Inc(P);
            if P^ = '|' then begin
                Inc(P);
                FToken := etCONCAT;
            end
            else
                DatabaseErrorFmt(SExprInvalidChar, [P^]);
        end;
    '*':
        begin
            Inc(P);
            FToken := etMUL;
        end;
    '/':
        begin
            Inc(P);
            FToken := etDIV;
      end;
    ',':
        begin
            Inc(P);
            FToken := etComma;
        end;
    #0:
        FToken := etEnd;
    else
        DatabaseErrorFmt(SExprInvalidChar, [P^]);
    end;
    FSourcePtr := P;
end;

function TOCIExpressionParser.ParseExpr: POCIExpressionNode;
begin
    Result := ParseExpr2;
    while TokenSymbolIs('OR') do begin
        NextToken;
        Result := FExpression.NewNode(enOperator, canOR, Unassigned,
            Result, ParseExpr2);
        GetScopeKind(Result, Result^.FLeft, Result^.FRight);
        Result^.FDataType := ftBoolean;
    end;
end;

function TOCIExpressionParser.ParseExpr2: POCIExpressionNode;
begin
    Result := ParseExpr3;
    while TokenSymbolIs('AND') do begin
        NextToken;
        Result := FExpression.NewNode(enOperator, canAND, Unassigned,
            Result, ParseExpr3);
        GetScopeKind(Result, Result^.FLeft, Result^.FRight);
        Result^.FDataType := ftBoolean;
    end;
end;

function TOCIExpressionParser.ParseExpr3: POCIExpressionNode;
begin
    if TokenSymbolIs('NOT') then begin
        NextToken;
        Result := FExpression.NewNode(enOperator, canNOT, Unassigned,
            ParseExpr4, nil);
        Result^.FDataType := ftBoolean;
    end
    else
        Result := ParseExpr4;
    GetScopeKind(Result, Result^.FLeft, Result^.FRight);
end;

function TOCIExpressionParser.ParseExpr4: POCIExpressionNode;
const
    Operators: array[etEQ..etLT] of TOCIExpressionOperator = (
        canEQ, canNE, canGE, canLE, canGT, canLT);
var
    Operator: TOCIExpressionOperator;
    Left, Right: POCIExpressionNode;
begin
    Result := ParseExpr5(False);
    if (FToken in [etEQ..etLT, etLIKE, etNOTLIKE, etISNULL, etISNOTNULL,
                   etIN, etNOTIN, etBETWEEN, etNOTBETWEEN]) then begin
        case FToken of
        etEQ..etLT:
            Operator := Operators[FToken];
        etLIKE:
            Operator := canLIKE;
        etNOTLIKE:
            Operator := canNOTLIKE;
        etISNULL:
            Operator := canISBLANK;
        etISNOTNULL:
            Operator := canNOTBLANK;
        etIN:
            Operator := canIN;
        etNOTIN:
            Operator := canNOTIN;
        etBETWEEN:
            Operator := canBETWEEN;
        etNOTBETWEEN:
            Operator := canNOTBETWEEN;
        else
            Operator := canNOTDEFINED;
        end;
        NextToken;
        Left := Result;
        if Operator in [canIN, canNOTIN] then begin
            if FToken <> etLParen then
                DatabaseErrorFmt(SExprNoLParen, [TokenName]);
            NextToken;
            Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, nil);
            Result.FDataType := ftBoolean;
            if FToken <> etRParen then begin
                Result.FArgs := TList.Create;
                repeat
                    Right := ParseExpr;
                    if IsTemporal(Left.FDataType) then
                        Right.FDataType := Left.FDataType;
                    Result.FArgs.Add(Right);
                    if (FToken <> etComma) and (FToken <> etRParen) then
                        DatabaseErrorFmt(SExprNoRParenOrComma, [TokenName]);
                    if FToken = etComma then
                        NextToken;
                until (FToken = etRParen) or (FToken = etEnd);
                if FToken <> etRParen then
                    DatabaseErrorFmt(SExprNoRParen, [TokenName]);
                NextToken;
            end
            else
                DatabaseError(SExprEmptyInList);
        end
        else if Operator in [canBETWEEN, canNOTBETWEEN] then begin
            Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, nil);
            Result.FDataType := ftBoolean;
            Result.FArgs := TList.Create;
            Result.FArgs.Add(ParseExpr5(False));
            if TokenSymbolIs('AND') then
                NextToken
            else
                DatabaseErrorFmt(SExprExpected, ['AND']);
            Result.FArgs.Add(ParseExpr5(False));
        end
        else begin
            if Operator in [canEQ, canNE, canGE, canLE, canGT, canLT] then begin
                Right := ParseExpr5(True);
                Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, Right);
            end
            else if Operator in [canLike, canNotLike] then begin
                Right := ParseExpr5(False);
                Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, Right);
                if TokenSymbolIs('ESCAPE') then begin
                    NextToken;
                    Result.FArgs := TList.Create;
                    Result.FArgs.Add(ParseExpr5(False));
                end;
            end
            else begin
                Right := nil;
                Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, Right);
            end;
            if Right <> nil then begin
                if (Left^.FKind = enField) and (Right^.FKind = enConst) then
                    Right^.FDataType := Left^.FDataType
                else if (Right^.FKind = enField) and (Left^.FKind = enConst) then
                    Left^.FDataType := Right^.FDataType
            end;
            if (Left^.FDataType in [ftBLOB, ftMEMO]) and (Operator = canLIKE) then begin
                if Right^.FKind = enConst then
                    Right^.FDataType := ftString;
            end
            else if (Operator <> canISBLANK) and (Operator <> canNOTBLANK)
                    and ((Left^.FDataType in [ftBLOB, ftMEMO, ftVarBytes]) or
                         ((Right <> nil) and (Right^.FDataType in [ftBLOB, ftMEMO, ftVarBytes]))) then
                DatabaseError(SExprTypeMis);
            Result.FDataType := ftBoolean;
            if Right <> nil then begin
                if IsTemporal(Left.FDataType) and (Right.FDataType = ftString) then
                    Right.FDataType := Left.FDataType
                else if IsTemporal(Right.FDataType) and (Left.FDataType = ftString) then
                    Left.FDataType := Right.FDataType;
            end;
            GetScopeKind(Result, Left, Right);
        end;
    end;
end;

function TOCIExpressionParser.ParseExpr5(AEnableList: Boolean): POCIExpressionNode;
const
    OperatorsList: array[etANY..etALL] of TOCIExpressionOperator = (
        canANY, canANY, canALL);
    Operators: array[etCONCAT..etDIV] of TOCIExpressionOperator = (
        canCONCAT, canADD, canSUB, canMUL, canDIV);
var
    Operator: TOCIExpressionOperator;
    Left, Right: POCIExpressionNode;
begin
    if FToken in [etANY, etSOME, etALL] then begin
        if not AEnableList then
            DatabaseErrorFmt(SExprExpected, [TokenName]);
        Operator := OperatorsList[FToken];
        NextToken;
        if FToken <> etLParen then
            DatabaseErrorFmt(SExprNoLParen, [TokenName]);
        NextToken;
        if FToken = etRParen then
            DatabaseError(SExprEmptyInList);
        Result := FExpression.NewNode(enOperator, Operator, Unassigned, nil, nil);
        Result.FArgs := TList.Create;
        repeat
            Result.FArgs.Add(ParseExpr);
            if (FToken <> etComma) and (FToken <> etRParen) then
                DatabaseErrorFmt(SExprNoRParenOrComma, [TokenName]);
            if FToken = etComma then
                NextToken;
        until (FToken = etRParen) or (FToken = etEnd);
        if FToken <> etRParen then
            DatabaseErrorFmt(SExprNoRParen, [TokenName]);
        NextToken;
        Result.FDataType := POCIExpressionNode(Result.FArgs.Items[0])^.FDataType;
        // ??? Result.FScopeKind := POCIExpressionNode(Result.FArgs.Items[0])^.FScopeKind;
    end
    else begin
        Result := ParseExpr6;
        while FToken in [etCONCAT, etADD, etSUB] do begin
            if not (poExtSyntax in FParserOptions) then
                DatabaseError(SExprNoArith);
            Operator := Operators[FToken];
            Left := Result;
            NextToken;
            Right := ParseExpr6;
            Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, Right);
            TypeCheckArithOp(Result);
            GetScopeKind(Result, Left, Right);
        end;
    end;
end;

function TOCIExpressionParser.ParseExpr6: POCIExpressionNode;
const
    Operators: array[etCONCAT..etDIV] of TOCIExpressionOperator = (
        canCONCAT, canADD, canSUB, canMUL, canDIV);
var
    Operator: TOCIExpressionOperator;
    Left, Right: POCIExpressionNode;
begin
    Result := ParseExpr7;
    while FToken in [etMUL, etDIV] do begin
        if not (poExtSyntax in FParserOptions) then
            DatabaseError(SExprNoArith);
        Operator := Operators[FToken];
        Left := Result;
        NextToken;
        Right := ParseExpr7;
        Result := FExpression.NewNode(enOperator, Operator, Unassigned, Left, Right);
        TypeCheckArithOp(Result);
        GetScopeKind(Result, Left, Right);
    end;
end;

function TOCIExpressionParser.ParseExpr7: POCIExpressionNode;
var
    FuncName: string;
    FuncInd: Integer;
begin
    case FToken of
    etSymbol:
        begin
            FuncInd := TokenSymbolIsFunc(FTokenString);
            if (poExtSyntax in FParserOptions) and (FuncInd <> -1) then begin
                FuncName := FTokenString;
                if not NextTokenIsLParen then begin
                    Result := FExpression.NewNode(enFunc, canNOTDEFINED, FuncName,
                        nil, nil, FuncInd);
                    NextToken;
                end
                else begin
                    NextToken;
                    if FToken <> etLParen then
                        DatabaseErrorFmt(SExprNoLParen, [TokenName]);
                    NextToken;
                    if (CompareText(FuncName,'count') = 0) and (FToken = etMUL) then begin
                        FuncName := 'COUNT(*)';
                        NextToken;
                    end;
                    Result := FExpression.NewNode(enFunc, canNOTDEFINED, FuncName,
                        nil, nil, FuncInd);
                    if FToken <> etRParen then begin
                        Result.FArgs := TList.Create;
                        repeat
                            Result.FArgs.Add(ParseExpr);
                            if (FToken <> etComma) and (FToken <> etRParen) then
                                DatabaseErrorFmt(SExprNoRParenOrComma, [TokenName]);
                            if FToken = etComma then
                                NextToken;
                        until (FToken = etRParen) or (FToken = etEnd);
                    end;
                end;
                GetFuncResultInfo(Result);
            end
            else if TokenSymbolIs('NULL') then begin
                Result := FExpression.NewNode(enConst, canNOTDEFINED, System.Null, nil, nil);
                Result.FScopeKind := skConst;
            end
            else if TokenSymbolIs(FStrTrue) then begin
                Result := FExpression.NewNode(enConst, canNOTDEFINED, 1, nil, nil);
                Result.FScopeKind := skConst;
            end
            else if TokenSymbolIs(FStrFalse) then begin
                Result := FExpression.NewNode(enConst, canNOTDEFINED, 0, nil, nil);
                Result.FScopeKind := skConst;
            end
            else begin
                Result := FExpression.NewNode(enField, canNOTDEFINED, FTokenString, nil, nil);
                Result.FScopeKind := skField;
            end;
        end;
    etName:
        begin
            Result := FExpression.NewNode(enField, canNOTDEFINED, FTokenString, nil, nil);
            Result.FScopeKind := skField;
        end;
    etLiteral:
        begin
            Result := FExpression.NewNode(enConst, canNOTDEFINED, FTokenString, nil, nil);
            if FNumericLit then
                Result^.FDataType := ftFloat
            else
                Result^.FDataType := ftString;
            Result.FScopeKind := skConst;
        end;
    etLParen:
        begin
            NextToken;
            Result := ParseExpr;
            if FToken <> etRParen then
                DatabaseErrorFmt(SExprNoRParen, [TokenName]);
        end;
    else
        DatabaseErrorFmt(SExprExpected, [TokenName]);
        Result := nil;
    end;
    NextToken;
end;

procedure TOCIExpressionParser.GetScopeKind(Root, Left, Right: POCIExpressionNode);
begin
    if (Left = nil) and (Right = nil) then
        Exit;
    if Right = nil then begin
        Root.FScopeKind := Left.FScopeKind;
        Exit;
    end;
    if ((Left^.FScopeKind = skField) and (Right^.FScopeKind = skAgg))
       or ((Left^.FScopeKind = skAgg) and (Right^.FScopeKind = skField)) then
        DatabaseError(SExprBadScope);
    if (Left^.FScopeKind = skConst) and (Right^.FScopeKind = skConst) then
        Root^.FScopeKind := skConst
    else if (Left^.FScopeKind = skAgg) or (Right^.FScopeKind = skAgg) then
        Root^.FScopeKind := skAgg
    else if (Left^.FScopeKind = skField) or (Right^.FScopeKind = skField) then
        Root^.FScopeKind := skField;
end;

procedure TOCIExpressionParser.GetFuncResultInfo(Node : POCIExpressionNode);
var
    i, n: Integer;
    pFuncDesc: TOCIFuncDesc;
begin
    i := TokenSymbolIsFunc(Node^.FData);
    if (Node^.FArgs = nil) then
        n := 0
    else
        n := Node^.FArgs.Count;
    pFuncDesc := GetFuncDesc(i);
    if (pFuncDesc.FArgMin > n) or (pFuncDesc.FArgMax < n) then
        DatabaseError(SExprTypeMis);
    if pFuncDesc.FDataType = ftUnknown then
        Node^.FDataType := POCIExpressionNode(Node^.FArgs.
            Items[pFuncDesc.FDataTypeArg])^.FDataType
    else
        Node^.FDataType := pFuncDesc.FDataType;
    if pFuncDesc.FScopeKind = skUnknown then
        Node^.FScopeKind := POCIExpressionNode(Node^.FArgs.
            Items[pFuncDesc.FScopeKindArg])^.FScopeKind
    else
        Node^.FScopeKind := pFuncDesc.FScopeKind;
end;

function TOCIExpressionParser.TokenName: string;
begin
    if FSourcePtr = FTokenPtr then
        Result := SExprNothing
    else begin
        SetString(Result, FTokenPtr, FSourcePtr - FTokenPtr);
        Result := '''' + Result + '''';
    end;
end;

function TOCIExpressionParser.TokenSymbolIs(const S: string): Boolean;
begin
    Result := (FToken = etSymbol) and (CompareText(FTokenString, S) = 0);
end;

function TOCIExpressionParser.TokenSymbolIsFunc(const S: string): Integer;
begin
    if not OCIFuncs.Find(UpperCase(S), Result) then
        Result := -1;
end;

procedure TOCIExpressionParser.TypeCheckArithOp(Node: POCIExpressionNode);
begin
    with Node^ do begin
        if IsNumeric(FLeft.FDataType) and IsNumeric(FRight.FDataType) then
            FDataType := ftFloat
        else if (FLeft.FDataType = ftString) and (FRight.FDataType = ftString) and
                ((FOperator = canADD) or (FOperator = canCONCAT)) then
            FDataType := ftString
        else if IsTemporal(FLeft.FDataType) and IsNumeric(FRight.FDataType) and
                (FOperator = canADD) then
            FDataType := ftDateTime
        else if IsTemporal(FLeft.FDataType) and IsNumeric(FRight.FDataType) and
                (FOperator = canSUB) then
            FDataType := FLeft.FDataType
        else if IsTemporal(FLeft.FDataType) and IsTemporal(FRight.FDataType) and
                (FOperator = canSUB) then
            FDataType := ftFloat
        else if (FLeft.FDataType = ftString) and IsTemporal(FRight.FDataType) and
                (FOperator = canSUB) then begin
            FLeft.FDataType := FRight.FDataType;
            FDataType := ftFloat;
        end
        else if (FLeft.FDataType = ftString) and IsNumeric(FRight.FDataType)and
                (FLeft.FKind = enConst) then
            FLeft.FDataType := ftDateTime
        else
            DatabaseError(SExprTypeMis);
    end;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIFilter

destructor TOCIFilter.Destroy;
begin
    if (Name = SDefaultFltName) and (DataSet <> nil) then begin
        DataSet.Filtered := False;
        DataSet.Filter := '';
        DataSet.FilterOptions := [];
    end;
    UnPrepare;
    inherited Destroy;
end;

function TOCIFilter.GetDisplayName: string;
begin
    if Name <> '' then
        Result := Name
    else
        Result := inherited GetDisplayName;
end;

procedure TOCIFilter.Assign(AValue: TPersistent);
begin
    if AValue is TOCIFilter then
        try
            if Collection <> nil then
                Collection.BeginUpdate;
            Active := False;
            Name := TOCIFilter(AValue).Name;
            Text := TOCIFilter(AValue).Text;
            Options := TOCIFilter(AValue).Options;
            Active := TOCIFilter(AValue).Active;
        finally
            if Collection <> nil then
                Collection.EndUpdate;
        end
    else
        inherited Assign(AValue);
end;

procedure TOCIFilter.Changed(ACanCache: Boolean);
begin
    if not (ACanCache and (DataSet <> nil) and DataSet.Active) then
        UnPrepare;
    inherited Changed(False);
end;

procedure TOCIFilter.Prepare;
var
    parser: TOCIExpressionParser;
begin
    if not FPrepared and Active and (DataSet <> nil) and
       ((Text <> '') or Assigned(FOnFilterRecord)) then begin
        if Text <> '' then begin
            parser := TOCIExpressionParser.Create;
            try
                FExpression := parser.BuildExpression(DataSet, Text, Options, [poExtSyntax], '');
            finally
                parser.Free;
            end;
        end;
        FPrepared := True;
    end;
end;

procedure TOCIFilter.UnPrepare;
begin
    if FPrepared then begin
        FPrepared := False;
        if FExpression <> nil then begin
            FExpression.Free;
            FExpression := nil;
        end;
    end;
end;

function TOCIFilter.Execute: Boolean;
begin
    try
        Result := True;
        if FExpression <> nil then
            Result := FExpression.Execute;
        if Result and Assigned(FOnFilterRecord) then
            FOnFilterRecord(DataSet, Result);
    except
        Application.HandleException(Self);
    end;
end;

procedure TOCIFilter.SetActive(const Value: Boolean);
begin
    if (FActive <> Value) and ((DataSet <> nil) or not Value) then begin
        FActive := Value;
        if (Name = SDefaultFltName) and (DataSet <> nil) then
            DataSet.Filtered := Value;
        Changed(True);
    end;
end;

procedure TOCIFilter.SetOptions(const Value: TFilterOptions);
begin
    if FOptions <> Value then begin
        FOptions := Value;
        if (Name = SDefaultFltName) and (DataSet <> nil) then
            DataSet.FilterOptions := Value;
        Changed(False);
    end;
end;

procedure TOCIFilter.SetText(const Value: String);
begin
    if FText <> Value then begin
        FText := Value;
        if (Name = SDefaultFltName) and (DataSet <> nil) then
            DataSet.Filter := Value;
        Changed(False);
    end;
end;

procedure TOCIFilter.SetFilterRecord(const Value: TFilterRecordEvent);
begin
    if @FOnFilterRecord <> @Value then begin
        FOnFilterRecord := Value;
        if (Name = SDefaultFltName) and (DataSet <> nil) then
            DataSet.OnFilterRecord := Value;
        Changed(False);
    end;
end;

function TOCIFilter.GetDataSet: TDataSet;
begin
    if not Assigned(Collection) then
        Result := nil
    else
        Result := TOciFilters(Collection).GetDataSet;
end;

function TOCIFilter.IsEqual(AValue: TOCIFilter): Boolean;
begin
    Result :=
        (Text = AValue.Text) and
        (Name = AValue.Name) and
        (Options = AValue.Options) and
        (Active = AValue.Active) and
        (@OnFilterRecord = @AValue.OnFilterRecord);
end;

{TODO
    We need some caching there, without it
    this code will work more than need
TODO}
procedure TOCIFilter.PrepareLookupFilter(const AKeyFields: String;
    const AKeyValues: Variant; AOptions: TLocateOptions);
var
    I: Integer;
    Node: POCIExpressionNode;
    FilterOptions: TFilterOptions;
    Fields: TList;
begin
    UnPrepare;
    if DataSet = nil then
        Exit;
    FilterOptions := [];
    if loCaseInsensitive in AOptions then
        Include(FilterOptions, foCaseInsensitive);
    if not (loPartialKey in AOptions) then
        Include(FilterOptions, foNoPartialCompare);
    Fields := TList.Create;
    try
        DataSet.GetFieldList(Fields, AKeyFields);
        FExpression := TOCIExpression.Create(DataSet, FilterOptions, [], '');
        FPrepared := True;
        if Fields.Count = 1 then
            FExpression.FRoot := FExpression.NewCompareNode(TField(Fields[0]),
                canEQ, AKeyValues)
        else
            for I := 0 to Fields.Count - 1 do begin
                Node := FExpression.NewCompareNode(TField(Fields[I]), canEQ,
                    AKeyValues[I]);
                if I = 0 then
                    FExpression.FRoot := Node
                else
                    FExpression.FRoot := FExpression.NewNode(enOperator, canAND,
                        Unassigned, FExpression.FRoot, Node);
            end;
    finally
        Fields.Free;
    end;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
{ TOCIFilters }

constructor TOCIFilters.Create(AOwner: TPersistent);
begin
    inherited Create(TOCIFilter);
    FOwner := AOwner;
end;

function TOCIFilters.GetOwner: TPersistent;
begin
    Result := FOwner;
end;

function TOCIFilters.GetDataSet: TDataSet;
begin
    if FOwner is TDataSet then
        Result := TDataSet(FOwner)
    else
        Result := nil;
end;

function TOCIFilters.IsEqual(Value: TOciFilters): Boolean;
var
    I: Integer;
begin
    Result := Count = Value.Count;
    if Result then
        for I := 0 to Count - 1 do begin
            Result := Items[I].IsEqual(Value.Items[I]);
            if not Result then
                Break;
        end
end;

function TOCIFilters.GetItem(AIndex: Integer): TOCIFilter;
begin
    Result := TOCIFilter(inherited Items[AIndex]);
end;

procedure TOCIFilters.SetItem(AIndex: Integer; AValue: TOCIFilter);
begin
    inherited SetItem(AIndex, AValue);
end;

function TOCIFilters.FindFilter(const Value: string): TOciFilter;
var
    I: Integer;
begin
    for I := 0 to Count - 1 do begin
        Result := Items[I];
        if AnsiCompareText(Result.Name, Value) = 0 then
            Exit;
    end;
    Result := nil;
end;

function TOCIFilters.FilterByName(const Value: string): TOciFilter;
begin
    Result := FindFilter(Value);
    if Result = nil then
        OCIDBErrorFmt(msgOCIFilterNotFound, [Value]);
end;

function TOCIFilters.GetDefaultFilter: TOCIFilter;
begin
    Result := FindFilter(SDefaultFltName);
    if Result = nil then begin
        Result := TOCIFilter(Add);
        Result.Name := SDefaultFltName;
    end;
end;

function TOCIFilters.GetLookupFilter: TOCIFilter;
begin
    Result := FindFilter(SLookupFltName);
    if Result = nil then begin
        Result := TOCIFilter(Add);
        Result.Name := SLookupFltName;
    end;
end;

function TOCIFilters.Execute: Boolean;
var
    i: Integer;
begin
    Result := True;
    for i := 0 to Count - 1 do
        if Items[i].Active then begin
            Result := Items[i].Execute;
            if not Result then
                Break;
        end;
end;

procedure TOCIFilters.Update(Item: TCollectionItem);
begin
    if (DataSet <> nil) and (DataSet is TOCIDataSet) then
        __TOCIDataSet(DataSet).UpdateFiltering(True);
end;

function TOCIFilters.GetAnyActive: Boolean;
var
    i: Integer;
begin
    Result := False;
    for i := 0 to Count - 1 do
        if Items[i].Active then begin
            Result := True;
            Exit;
        end;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Initialization && finalization

var
    i: Integer;

initialization
    OCIFuncs := TStringList.Create;
    OCIFuncs.Capacity := 100;
    OCIFuncs.Sorted := True;
    OCIFuncs.Duplicates := dupError;

    OCIAddFunc('UPPER',     skUnknown,  0,  ftString,   -1,  1,  1,  FunUpper);
    OCIAddFunc('LOWER',     skUnknown,  0,  ftString,   -1,  1,  1,  FunLower);
    OCIAddFunc('SUBSTRING', skUnknown,  0,  ftString,   -1,  2,  3,  FunSubstring);
    OCIAddFunc('TRIM',      skUnknown,  0,  ftString,   -1,  1,  2,  FunTrim);
    OCIAddFunc('TRIMLEFT',  skUnknown,  0,  ftString,   -1,  1,  2,  FunTrimLeft);
    OCIAddFunc('TRIMRIGHT', skUnknown,  0,  ftString,   -1,  1,  2,  FunTrimRight);
    OCIAddFunc('YEAR',      skUnknown,  0,  ftSmallInt, -1,  1,  1,  FunYear);
    OCIAddFunc('MONTH',     skUnknown,  0,  ftSmallInt, -1,  1,  1,  FunMonth);
    OCIAddFunc('DAY',       skUnknown,  0,  ftSmallInt, -1,  1,  1,  FunDay);
    OCIAddFunc('HOUR',      skUnknown,  0,  ftSmallInt, -1,  1,  1,  FunHour);
    OCIAddFunc('MINUTE',    skUnknown,  0,  ftSmallInt, -1,  1,  1,  FunMinute);
    OCIAddFunc('SECOND',    skUnknown,  0,  ftSmallInt, -1,  1,  1,  FunSecond);
    OCIAddFunc('GETDATE',   skConst,   -1,  ftDateTime, -1,  0,  0,  FunGetDate);
    OCIAddFunc('DATE',      skUnknown,  0,  ftDateTime, -1,  1,  1,  FunDate);
    OCIAddFunc('TIME',      skUnknown,  0,  ftDateTime, -1,  1,  1,  FunTime);
    OCIAddFunc('SUM',       skAgg,     -1,  ftFloat,    -1,  1,  1,  nil);
    OCIAddFunc('MIN',       skAgg,     -1,  ftUnknown,   0,  1,  1,  nil);
    OCIAddFunc('MAX',       skAgg,     -1,  ftUnknown,   0,  1,  1,  nil);
    OCIAddFunc('AVG',       skAgg,     -1,  ftFloat,    -1,  1,  1,  nil);
    OCIAddFunc('COUNT',     skAgg,     -1,  ftInteger,  -1,  0,  0,  nil);
    // ---------------------------------------------
    // Oracle's
    // Numeric
    OCIAddFunc('ABS',       skUnknown,  0,  ftFloat,    -1,  1,  1,  FunAbs);
    OCIAddFunc('CEIL',      skUnknown,  0,  ftInteger,  -1,  1,  1,  FunCeil);
    OCIAddFunc('COS',       skUnknown,  0,  ftFloat,    -1,  1,  1,  FunCos);
    OCIAddFunc('COSH',      skUnknown,  0,  ftFloat,    -1,  1,  1,  FunCosh);
    OCIAddFunc('EXP',       skUnknown,  0,  ftFloat,    -1,  1,  1,  FunExp);
    OCIAddFunc('FLOOR',     skUnknown,  0,  ftInteger,  -1,  1,  1,  FunFloor);
    OCIAddFunc('LN',        skUnknown,  0,  ftFloat,    -1,  1,  1,  FunLn);
    OCIAddFunc('LOG',       skUnknown,  0,  ftFloat,    -1,  2,  2,  FunLog);
    OCIAddFunc('MOD',       skUnknown,  0,  ftInteger,  -1,  2,  2,  FunMod);
    OCIAddFunc('POWER',     skUnknown,  0,  ftFloat,    -1,  2,  2,  FunPower);
    OCIAddFunc('ROUND',     skUnknown,  0,  ftFloat,    -1,  1,  2,  FunRound);
    OCIAddFunc('SIGN',      skUnknown,  0,  ftInteger,  -1,  1,  1,  FunSign);
    OCIAddFunc('SIN',       skUnknown,  0,  ftFloat,    -1,  1,  1,  FunSin);
    OCIAddFunc('SINH',      skUnknown,  0,  ftFloat,    -1,  1,  1,  FunSinh);
    OCIAddFunc('SQRT',      skUnknown,  0,  ftFloat,    -1,  1,  1,  FunSqrt);
    OCIAddFunc('TAN',       skUnknown,  0,  ftFloat,    -1,  1,  1,  FunTan);
    OCIAddFunc('TANH',      skUnknown,  0,  ftFloat,    -1,  1,  1,  FunTanh);
    OCIAddFunc('TRUNC',     skUnknown,  0,  ftFloat,    -1,  1,  2,  FunTrunc);
    // Strings, result string
    OCIAddFunc('CHR',       skUnknown,  0,  ftString,   -1,  1,  1,  FunChr);
    OCIAddFunc('CONCAT',    skUnknown,  0,  ftString,   -1,  2,  2,  FunConcat);
    OCIAddFunc('INITCAP',   skUnknown,  0,  ftString,   -1,  1,  1,  FunInitCap);
    OCIAddFunc('LPAD',      skUnknown,  0,  ftString,   -1,  2,  3,  FunLPad);
    OCIAddFunc('LTRIM',     skUnknown,  0,  ftString,   -1,  1,  2,  FunTrimLeft);
    OCIAddFunc('REPLACE',   skUnknown,  0,  ftString,   -1,  2,  3,  FunReplace);
    OCIAddFunc('RPAD',      skUnknown,  0,  ftString,   -1,  2,  3,  FunRPad);
    OCIAddFunc('RTRIM',     skUnknown,  0,  ftString,   -1,  1,  2,  FunTrimRight);
    OCIAddFunc('SOUNDEX',   skUnknown,  0,  ftString,   -1,  1,  1,  FunSoundex);
    OCIAddFunc('SUBSTR',    skUnknown,  0,  ftString,   -1,  2,  3,  FunSubstring);
    OCIAddFunc('TRANSLATE', skUnknown,  0,  ftString,   -1,  3,  3,  FunTranslate);
    // Strings, result integer
    OCIAddFunc('ASCII',     skUnknown,  0,  ftInteger,  -1,  1,  1,  FunAscii);
    OCIAddFunc('INSTR',     skUnknown,  0,  ftInteger,  -1,  2,  4,  FunInstr);
    OCIAddFunc('LENGTH',    skUnknown,  0,  ftInteger,  -1,  1,  1,  FunLength);
    // Date
    OCIAddFunc('ADD_MONTHS',skUnknown,  0,  ftDateTime, -1,  2,  2,  FunAddMonths);
    OCIAddFunc('LAST_DAY',  skUnknown,  0,  ftDateTime, -1,  1,  1,  FunLastDay);
    OCIAddFunc('MONTHS_BETWEEN',
                            skUnknown,  0,  ftInteger,  -1,  2,  2,  FunMonthsBW);
    OCIAddFunc('SYSDATE',   skConst,   -1,  ftDateTime, -1,  1,  1,  FunGetDate);
    // Convert
    OCIAddFunc('TO_CHAR',   skUnknown,  0,  ftString,   -1,  1,  2,  FunToChar);
    OCIAddFunc('TO_DATE',   skUnknown,  0,  ftDateTime, -1,  1,  2,  FunToDate);
    OCIAddFunc('TO_NUMBER', skUnknown,  0,  ftFloat,    -1,  1,  2,  FunToNumber);
    // Others
    OCIAddFunc('DECODE',    skUnknown,  2,  ftUnknown,   2,  3,  MAXINT,  FunDecode);
    OCIAddFunc('IIF',       skUnknown,  1,  ftUnknown,   1,  3,  3,       FunIIF);
    OCIAddFunc('IF',        skUnknown,  1,  ftUnknown,   1,  3,  3,       FunIIF);
    OCIAddFunc('NVL',       skUnknown,  0,  ftUnknown,   0,  2,  2,       FunNvl);
    OCIAddFunc('GREATEST',  skUnknown,  0,  ftUnknown,   0,  1,  MAXINT,  FunGreatest);
    OCIAddFunc('LEAST',     skUnknown,  0,  ftUnknown,   0,  1,  MAXINT,  FunLeast);

finalization
    for i := 0 to OCIFuncs.Count - 1 do
        OCIFuncs.Objects[i].Free;
    OCIFuncs.Free;

end.
